#!/usr/bin/python
'''
Module including functionality that allows to elaborate and extract useful 
information from pcz files.						
'''

from __future__ import absolute_import, print_function, division

import sys
import logging as log

import numpy as np
import math

import mdtraj as mdt
from MDPlus.analysis import pca
import tempfile

def pczdump(args):

    listLong = ['--input', '--evec', '--fluc', '--evals', '--avg', '--coll', 
                '--proj', '--info', '--verbosity', '--output', '--rms', 
                '--anim' ]
    listShort = ['-i', '-e', '-f', '-l', '-a', '-c', '-p', '-n', '-v', '-o', 
                 '-r']

    numArgs = 0
    for i in range(1, len(sys.argv)):
        if (sys.argv[i] in listShort and 
            listLong[listShort.index(sys.argv[i])] in sys.argv):
            log.error('Please, use either the long or the short form of '
                      + 'an option but never both! Try again!')
            sys.exit(-1)

    listUniqOpts = ['--evec', '--fluc', '--evals', '--avg', '--coll', 
                    '--proj', '--info', '--rms', '--anim', '-e', '-f',
                    '-l', '-a', '-c', '-p', '-n', '-r', '-m']
    for i in range(1, len(sys.argv)):
        if sys.argv[i] in listUniqOpts:
            numArgs += 1
            if numArgs == 2:
                log.error('Please ask for an option at a time from pyPczdump. '
                          + 'You have currently been asking for more than one '
                          + 'option! Try again!''')
                sys.exit(-1)

    if args.verbosity:
        if args.verbosity > 1:
            log.basicConfig(format="%(levelname)s: %(message)s", 
                            level=log.DEBUG)
        elif args.verbosity == 1:
            log.basicConfig(format="%(levelname)s: %(message)s", level=log.INFO)
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    if (args.input is None):
        log.error('')
        log.error('All or any of the mandatory command line arguments is '
                  + 'missing. The correct usage of pypczdump should be:')
        log.error('pyPczdump -i|--input <input-file> [optional arguments]')
        log.error('')
        log.error('Type "pyPczdump -h" or "pyPczdump --help" for further '
                  + 'details.')
        log.error('')
        sys.exit(-1)

    try:
        pcz_instance = pca.load(args.input)
    except IOError:
        exit(1)
    if args.evec is not None:
        if int(args.evec) == 0:
            log.error('pyPczdump used 1-based indexing, not zero-based.')
            sys.exit(-1)
        e = pcz_instance.evecs[int(args.evec) - 1]
        if args.output is not None:
            np.savetxt(args.output, np.column_stack((e,)))
        else:
            log.info("Values of eigenvector %d follow:\n" % int(args.evec))
            for i in e:
                print(i)

    elif args.proj is not None:
        if '-' in args.proj:
            p1 = int(args.proj[:args.proj.index('-')])
            p2 = int(args.proj[args.proj.index('-') + 1:]) 
            if p1 == 0:
                log.error('pyPczdump uses 1-based indexing, not zero-based.')
                sys.exit(-1)
            p = pcz_instance.projs[p1 - 1:p2].T
        else:
            if int(args.proj) == 0:
                log.error('pyPczdump uses 1-based indexing, not zero-based.')
                sys.exit(-1)
            p = pcz_instance.projs[int(args.proj) - 1]
        if args.output is not None:
            np.savetxt(args.output, np.column_stack((p,)))
        else:
            log.info('Values of the projections of eigenvector(s)'
                     + ' %s follow:\n' % args.proj)
            for i in p:
                if len(p.shape) > 1:
                    fmt = '%f ' * len(i) + '\n'
                    print(fmt % tuple(i))
                else:
                    print(i)
    elif args.evals is True:
        evs = pcz_instance.evals
        if args.output is not None:
            np.savetxt(args.output, np.column_stack((evs,)))
        else:
            log.info("The list of eigenvalues follows:\n")
            for i in evs:
                print(i)
    elif args.info is True:
        if pcz_instance.title is not None:
            log.info("Basic information on the compressed file:\n")
            print("%s\n" % pcz_instance.title)
        print("The PCZ file format version is %s\n" % pcz_instance.version)
        print("The number of atoms is %d\n" % pcz_instance.n_atoms)
        print("The number of frames is %d.\n" % pcz_instance.n_frames)
        print("The number of eigenvectors is %d.\n" % pcz_instance.n_vecs)
        print("The variance captured inside this file is %.2f%%.\n" % (
              100 * pcz_instance.evals.sum() / pcz_instance.totvar))
    elif args.fluc is not None:
        if int(args.fluc) == 0:
            log.error('pyPczdump uses 1-based indexing, not zero-based.')
            sys.exit(-1)
        evec = pcz_instance.evecs[int(args.fluc) - 1].reshape((-1, 3))
        fluc = np.sqrt((evec * evec).sum(axis=1))

        if args.output is not None:
            np.savetxt(args.output, np.column_stack((fluc,)))
        else:
            log.info('Values of the fluctuations associated with '
                      + 'eigenvector %d follow:\n' % int(args.fluc))
            for i in fluc:
                print(i)
    elif args.avg is True:
        pcz_instance = pca.load(args.input)
        avg = pcz_instance.refxyz

        r = (pcz_instance.evecs.T *
             pcz_instance.evals).reshape(pcz_instance.n_atoms, -1)
        rmsf = np.sqrt((r * r).sum(axis=1))
        if args.output is None:
            out = tempfile.NamedTemporaryFile(suffix='.pdb').name
        else:
            out = args.output
        u = mdt.Trajectory(avg * 0.1, pcz_instance.topology)
        u.save(out, bfactors=rmsf)

        if args.output is None:
            with open(out, 'rb') as f:
                for line in f:
                    print(line[:-1].decode('utf-8'))

    elif args.coll is True:
        col = np.zeros((pcz_instance.n_vecs))
        for i in range(pcz_instance.n_vecs):
            e = pcz_instance.evecs[i]
            for j in range(0, pcz_instance.n_atoms * 3, 3):
                r2 = e[j] * e[j] + e[j + 1] * e[j + 1] + e[j + 2] * e[j + 2]
                col[i] = col[i] - r2 * math.log(r2)
        col = np.exp(col) / pcz_instance.n_atoms
        if args.output is None:
            log.info('Collectivity metric values K for each eigenvector '
                     + 'follow.:\n')
            log.info('Modes producing most of the collective motion in the '
                     + 'system will have high K values.\n')
            for i in col:
                print(i)
        else:
            np.savetxt(args.output, np.column_stack((col,)))

    elif args.rms is not None:
        if ((int(args.rms) > 0) and (int(args.rms) <= pcz_instance.n_frames)):
            s2 = pcz_instance.scores(int(args.rms) - 1)
        rmsd = np.zeros((pcz_instance.n_frames))
        for i in range(pcz_instance.n_frames):
            rmsd[i] = 0.0
            s1 = pcz_instance.scores(i)
            if ((int(args.rms) > 0) and (int(args.rms) <= pcz_instance.n_frames)):
                s1 = s1 - s2
            rmsd[i] = np.sum(s1 * s1) / pcz_instance.n_atoms

        if args.output is not None:
            np.savetxt(args.output, np.column_stack((rmsd,)))
        else:
            for i in rmsd:
                print(np.sqrt(i))

    elif args.anim is not None:
        avg = pcz_instance.refxyz
        if int(args.anim) == 0:
            log.error('pyPczdump uses 1-based indexing, not zero-based.')
            sys.exit(-1)
        evec = pcz_instance.evecs[int(args.anim) - 1]
        proj = pcz_instance.projs[int(args.anim) - 1]

        ev = evec.reshape((-1, 3))
        fluc = np.sqrt((ev * ev).sum(axis=1))

        rmin = np.min(proj)
        rmax = np.max(proj)
        rinc = (rmax - rmin) * 0.1
        proj = np.zeros(20)
        for i in range(1, 6, 1):
            proj[i] = proj[i - 1] + rinc
        for i in range(6, 16, 1):
            proj[i] = proj[i - 1] - rinc
        for i in range(16, 20, 1):
            proj[i] = proj[i - 1] + rinc

        if args.output is None:
            out = tempfile.NamedTemporaryFile(suffix='.pdb').name
        else:
            out = args.output

        trj = []
        for j in range(20):
            trj.append(avg.flatten() + evec * proj[j])

        trj = np.array(trj) * 0.1 # convert to nanometers
        trj = trj.reshape((20, -1, 3))
        u = mdt.Trajectory(trj, pcz_instance.topology)
        u.save(out, bfactors=fluc)

        if args.output is None:
            with open(out, 'rb') as f:
                for line in f:
                    print(line[:-1].decode('utf-8'))




#!/usr/bin/env python
"""
 The python version of pcazip!!

 Reproduces the functionality of the old fortran- and C-based versions.

 In essence its wraps a simple set of procedures provided by two modules:
       'cofasu' - trajectory file handling
       'pcz'    - PCA analysis

 Stripped down to the bare essentials, the complete procedure is:

 >>> from MDPlus.analysis.pca import Pcz
 >>> from MDPlus.core import Fasu, Cofasu
 >>>
 >>> f = Fasu('topology.top','trajectory.traj')
 >>> c = Cofasu(f)
 >>> p = Pcz(c)
 >>> p.write('compressed.pcz')

 Everything else is basically analysing and sanity-checking the arguments
 given on the command line.
"""

from __future__ import absolute_import, print_function, division

# General python libraries import.
import os.path as op
import sys
import logging as log
import warnings
#warnings.simplefilter('ignore')

from MDPlus.analysis import pca
from MDPlus.core import Fasu, Cofasu
from MDPlus.utils import pdb2selection
import numpy as np
from time import time
import tempfile

'''
Begin by defining the little utility function that parses trajectory
filename strings that have the "trajfile(start:stop:step)" structure.
'''

def input_parse(infile):
    if "(" in infile:
        i = infile.find("(")
        if ")" in infile[i:]:
            j = infile.find(")")
            sel = infile[i + 1:j]
            nc = sel.count(':')
            if nc == 0:
                sel = sel + '::'
            elif nc == 1:
                sel = sel + ':'
            start = sel.partition(':')[0].partition(':')[0]
            stop = sel.partition(':')[2].partition(':')[0]
            step = sel.partition(':')[2].partition(':')[2]

            if not start.isdigit():
                start = 0
            else:
                start = int(start)
            if not stop.isdigit():
                stop = None
            else:
                stop = int(stop)
            if not step.isdigit():
                step = 1
            else:
                stop = int(stop)
            if start == 0:
                log.error('Use 1-based indices when defining trajectory slices')
                sys.exit(-1)
            return infile[:i], slice(start-1, stop, step)
        else:
            log.error('Malformed trajectory filename: {0}'.format(infile))
            sys.exit(-1)
    else:
        return infile, slice(0, None, 1)


#############################################################################
#                                                                           #
#                        PCAZIP main function (start)                       #
#                                                                           #
#############################################################################

def pcazip(args):

    # Time the complete run time
    time0start = time()

    if args.nompi:
        comm = None
        rank = 0
        size = 1
    else:
        try:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
            size = comm.Get_size()
        except ImportError:
            comm = None
            rank = 0
            size = 1


    if args.verbosity:
        if args.verbosity > 1:
            log.basicConfig(format="%(levelname)s: %(message)s", 
                            level=log.DEBUG)

        elif args.verbosity == 1:
            log.basicConfig(format="%(levelname)s: %(message)s", level=log.INFO)
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    if args.tests:
        from subprocess import call
        import os
        try:
            tdir = tempfile.mkdtemp()
            log.info('Creating temporary test directory {}'.format(tdir))
            here = os.getcwd()
            os.chdir(tdir)
            call(['wget','-q','-P', tdir,
                     'https://bitbucket.org/ramonbsc/pypcazip/downloads/test200.tar.gz'])
            call(['tar', 'xf', tdir + '/test200.tar.gz', "-C", tdir])
            os.chdir(tdir + '/test')
            call(['./run_tests.sh'])
            os.chdir(here)
            if (rank == 0):
                log.info('Please explore additional testing scripts and related trajectory files as necessary at '+op.expanduser(tdir + '/test'))
        except:
            if (rank == 0):
                log.error('')
                log.error('Error while trying to test the correct installation of the pyPcazip suite tools!')
                log.error('')        
            sys.exit(-1)
        sys.exit(0)
			
            
    '''
    Input filename and topology filename are mandatory. Hence a check on
    these two parameters should be performed:
    '''
    if (not ((args.input is None) ^ (args.album is None))) or (args.topology is None):
        if rank == 0:
            log.error('')
            log.error('All or any of the mandatory command line arguments is missing. The correct usage is:')
            log.error( 'python ./pcazip.py XOR[(-i|--input <input-file>),(-a|--album) <album-file>] -t|--topology <topology-file> [optional arguments]')
            log.error('')
            log.error('Type "python ./pcazip.py -h" or "python ./pcazip.py --help" for further details.')
            log.error('')
        sys.exit(-1)

    '''
    Multiple album files OR multiple trajectory files are permitted.
    The rule is that all the trajectory files in each album must be
    compatible with a single topology file. If more than one album file
    is specified, then either there should be one topology file that is
    applicable to ALL the album files, or there should be one topology file
    specified for each album file. Similar rules operate in cases where
    multiple trajectory files are specified on the command line: either one
    topology file common to all trajectory files, or one topology file per
    trajectory file, must be given. The same rules operate (independently)
    for the selection and masking options: either there should be one
    selection/mask that applies to ALL trajectory files or albums, or there
    should be one selection/mask for each trajectory file or album.

    Let's check that these rules are being followed. First for albums:
    '''
    if args.input is None:
        na = len(args.album)
        nt = len(args.topology)
        if args.selection is None:
            ns = 1
        else:
            ns = len(args.selection)
        if args.mask is not None:
            ns = max(ns, len(args.mask))
        if nt > 1 and nt != na:
            if rank == 0:
                log.error(("Number of topology files must be one,"
                           " or equal to the number of album files."))
            sys.exit(-1)
        if ns > 1 and ns != na:
            if rank == 0:
                log.error(("Number of masks/selections must be one,"
                           " or equal to the number of album files."))
            sys.exit(-1)
    else:
        # now for trajectories:
        na = len(args.input)
        nt = len(args.topology)
        if args.selection is None:
            ns = 1
        else:
            ns = len(args.selection)
        if args.mask is not None:
            ns = max(ns, len(args.mask))
        if nt > 1 and nt != na:
            if rank == 0:
                log.error(("Number of topology files must be one, or equal"
                           " to the number of trajectory files."))
            sys.exit(-1)
        if ns > 1 and ns != na:
            if rank == 0:
                log.error(("Number of masks/selections must be one, or equal"
                           " to the number of trajectory files."))
            sys.exit(-1)
    '''
        We can now build the key data structures.
        The data structures are:

        uniStr[]:              a list of albums a[], one per topology file.
        a[]:                   a list of trajectory specifiers (each of length 4)
        traj. specifier:       [topfile, trajfile, slice, filter] where trajfile
                               is a string containing the trajectory filename,
                               topfile is the appropriate topology file, slice
                               is a string that defines which snapshots in the
                               trajectory are to be included, using the
                               conventional start:stop:step syntax that e.g.
                               numpy uses to slice arrays, and filter is the atom
                               selection string (MDAnalysis format).

    '''
    uniStr = []
    if args.input is None:
        try:
            # There are one or more album files to process. Within an album,
            # all trajectory files will share the same topology file and filter
            # specification.
            for i in range(len(args.album)):
                log.debug('Reading album file {0}'.format(i))
                # sort out the selection string:
                if args.selection is None:
                    sel = 'all'
                else:
                    if len(args.selection) == 1:
                        sel = args.selection[0]
                    else:
                        sel = args.selection[i]
                if args.mask is not None:
                    if len(args.mask) == 1:
                        sel =  pdb2selection(args.mask[0])
                    else:
                        sel =  pdb2selection(args.mask[i])
                # sort out the topology file string:
                if len(args.topology) == 1:
                    top = args.topology[0]
                else:
                    top = args.topology[i]

                # Files opened in text-mode
                for input_str in open(args.album[i]):
                    # Here, we should figure out whether the input_str contains
                    #  a "\n" and in that case not append a "null"-name file
                    # that would trigger an error in further processing. We
                    # should do this step before calling this function
                    # from the point where we read the single line of the album.
                    if input_str != '\n':
                        # EOLs are converted to '\n'
                        input_str = input_str.rstrip('\n')
                        l = input_parse(input_str)
                        a = [top, l[0], l[1], sel]
                        uniStr.append(a)
        except IOError as e:
            if rank == 0:
                log.error("Problems while tried to process the album file.")
                log.error(("Check whether the album file does exist and whether"
                           " its name matches the name given in input.\n"))
                log.error("I/O error({0}): {1}".format(e.errno, e.strerror))
            sys.exit(-2)
    else:
        '''
        One or more trajectory files have ben specified by the user, rather
        than one or more album files.
        '''
        for i in range(len(args.input)):
            log.debug('Reading trajectory file {0}'.format(i))
            # sort out the selection string:
            if args.selection is None:
                sel = 'all'
            else:
                if len(args.selection) == 1:
                    sel = args.selection[0]
                else:
                    sel = args.selection[i]
            if args.mask is not None:
                if len(args.mask) == 1:
                    sel = pdb2selection(args.mask[0])
                else:
                    sel = pdb2selection(args.mask[i])
            # sort out the topology file string:
            if len(args.topology) == 1:
                top = args.topology[0]
            else:
                top = args.topology[i]

            input_str = args.input[i]
            l = input_parse(input_str)
            a = [top, l[0], l[1], sel]
            uniStr.append(a)

    # Now we can create the cofasu:
    f = []
    kwargs = {}
    if args.centre is not None:
        kwargs['centre'] = args.centre
        kwargs['pack_into_box'] = True
        if rank == 0:
            log.info('Will place group {0} at centre of box to fix jumps'.format(args.centre))
    #
    # To be nice to the user, we check if the netCDF4 module is
    # available, and if not make sure to trap attempts to load
    # AMBER .ncdf format trajectory files.
    #
    try:
        import netCDF4
        nonetCDF4 = False
    except ImportError:
        nonetCDF4 = True

    #Time reading/gathering of the trajectories in parallel
    if rank == 0:
        log.info('Reading trajectory files...')

    time1start = time()
    i = 0
    for a in uniStr:

        log.debug('Cofasu:{0} {1} {2} {3}'.format(a[0], a[1], a[2], a[3]))
        if op.splitext(a[1])[1].lower() == '.ncdf' and nonetCDF4:
            log.error('netcdf4-python with the netCDF4 and HDF5 libraries must be installed to read AMBER .ncdf files.\nSee installation instructions at https://code.google.com/p/mdanalysis/wiki/netcdf')
            exit(1)
        f.append(Fasu(a[1], top=a[0], frames=a[2], selection=a[3], **kwargs))
        i += 1
    try:
        cf = Cofasu(f, comm=comm)
    except(ValueError):
        if rank == 0:
            log.error('Can\'t compile trajectory files - inconsistent sizes?')
        raise
    time1end = time()
    if rank == 0:
        log.debug('Time to read trajectory files: {0:.2f} s'.format(time1end -
                  time1start))

    if args.trj_output is not None:
        cf.write(args.trj_output)
        if rank == 0:
            log.info('Wrote selected frames and atoms ' 
                     + 'to trajectory file {0}'.format(args.trj_output))

    if args.nopca is False:
        # run the pca analysis:
        if rank == 0:
            if args.weighted:
                log.info('Running mass-weighhted pca analysis')
            else:
                log.info('Running pca analysis')
        time2start = time()
        p = pca.fromtrajectory(cf, quality=float(args.quality), 
                               req_evecs=args.evecs,
                               weighted=args.weighted,
                               fastmethod=args.fast)
        time2end = time()
        if rank == 0:
            log.debug('Time for pcz-analysis: {0:.2f} s'.format(time2end -
                     time2start))
            log.info("Writing compressed trajectory")

        if args.output is not None:
            output_file = args.output
        else:
            # The input trajectory file is a mandatory argument and the check
            # on this has been done previously.
            dir = op.dirname(uniStr[0][0][1])
            base_out_compressed = op.basename(uniStr[0][0][1])
            name_out_compressed = op.splitext(base_out_compressed)[0]
            output_file = op.join(dir, name_out_compressed + "_output.pcz")

        if rank == 0:
            time_write_output_0 = time()
            p.write(output_file)
            time_write_output_1 = time()

        if args.pdb_out is not None:
            cf.write(args.pdb_out, cf[0])
        if rank == 0:
            totTime = time() - time0start
            log.debug('Time to write the output file: {0:.2f} s'.format(
                time_write_output_1 - time_write_output_0))
            log.debug('Total run time:: {0:.2f} s\n'.format(totTime))



from __future__ import absolute_import, print_function, division


def pib(coords, box):
    """
    Pack coordinates into periodic box.

    Args:
        coords: (N,3) numpy array of atom coordinates.
        box: (3,3) numpy array of box parameters.

    Returns:
        coords: (N,3) numpy array of packed coordinates (overwrites input).
    """
    import numpy as np

    import numpy as np

    assert len(coords.shape) == 2
    assert coords.shape[1] == 3
    assert box.shape == (3, 3)

    boxinv = np.zeros((3))
    boxinv[0] = 1.0 / box[0,0]
    boxinv[1] = 1.0 / box[1,1]
    boxinv[2] = 1.0 / box[2,2]

    for xyz in coords:
        s = np.floor(xyz[2] * boxinv[2])
        xyz[2] -= s * box[2,2]
        xyz[1] -= s * box[2,1]
        xyz[0] -= s * box[2,0]

        s = np.floor(xyz[1] * boxinv[1])
        xyz[1] -= s * box[1,1]
        xyz[0] -= s * box[1,0]

        s = np.floor(xyz[0] * boxinv[0])
        xyz[0] -= s * box[0,0]

    return coords
def pdb2selection(pdbfile):
    '''
    Create selection strings from pdbfile data.

    A little utility function to convert 'mask' pdb files into
    MDTraj-style selection strings. Basically it reads the second column
    (atom number) and uses it to construct 'index' selections. Runs
    of consecutive numbers are expressed in 'start to stop' form.

    Args:
        pdbfile (str): Name of a PDB file

    Returns:
        selection (str): MDtraj format selection string
    '''

    sel = ''
    i = 0
    j = 0
    with open(pdbfile, 'r') as f:
        for line in f:
            if line.find('ATOM') == 0 or line.find('HETATM') == 0:
                k = int(line.split()[1])
                # the next line catches the initialization process:
                if i == 0:
                    i = k
                    j = k - 1
                # are we in a run of consecutive numbers?:
                if k == j + 1:
                    j = k
                else:
                    # time to write out another selection:
                    sel = sel + ' index {0} to {1} or'.format(i, j)
                    i = k
                    j = k
                # end-of-file reached. Make sure last selection is included:
    if i > 0 and j > 0:
        sel = sel + ' index {0} to {1} or'.format(i, j)
    if len(sel) > 3:
        # remove the trailing ' or':
        sel = sel[:-3]
    return sel




#!/usr/bin/python

from __future__ import absolute_import, print_function, division
from six.moves import range

import sys
import os.path as op
import logging as log

import numpy as np

import mdtraj as mdt
from MDPlus.analysis import pca
import warnings
#warnings.simplefilter('ignore')

#######################################################
# MAIN FUNCTION
#######################################################

def pcaunzip(args):

    if args.verbosity:
        if args.verbosity > 1:
            log.basicConfig(format="%(levelname)s: %(message)s", 
                            level=log.DEBUG)
        elif args.verbosity == 1:
            log.basicConfig(format="%(levelname)s: %(message)s", level=log.INFO)
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    if (args.compressed is None ):
        log.error('')
        log.error(
            'All or any of the mandatory command line arguments is missing. '
             + 'The correct usage of pyPcaunzip should be:')
        log.error(
            'pyPcaunzip -c|--compressed <compressed-file>,  [optional arguments]')
        log.error('')
        log.error('Type "pyPcaunzip -h" or "pyPcaunzip --help" '
                  + 'for further details.')
        log.error('')
        sys.exit(-1)

    # If no name is provided for the output file, the extension of the pcz file is just changed into .dcd.
    if not args.output:
        dir = op.dirname(args.compressed)
        base = op.basename(args.compressed)
        name = op.splitext(base)[0]
        args.output = op.join(dir, name + ".dcd")
        if op.isfile(args.output):
            log.error('Output file {} already exists - will not be'
                      + ' overwritten'.format(arg.output))
            exit(1)

    try:
        import netCDF4
        nonetCDF4 = False
    except ImportError:
        nonetCDF4 = True

    ext = op.splitext(args.output)[1].lower()
    if ext == '.ncdf' and nonetCDF4:
        log.error('netcdf4-python with the netCDF4 and HDF5 libraries must '
                  + 'be installed to write AMBER .ncdf files.\nSee '
                  + 'https://code.google.com/p/mdanalysis/wiki/netcdf')
        exit(1)

    log.info("PCAunzipping")
    pfile = pca.load(args.compressed)
    if ext == '.xtc':
        scalefac = 0.1 # Gromacs works in nanometers.
    else:
        scalefac = 1.0
    regularize=False
    if args.regularize:
        regularize=True
    with mdt.open(args.output, 'w') as f:
        for ts_index in range(pfile.n_frames):
            f.write(pfile.frame(ts_index, regularize=regularize) * scalefac)




from __future__ import absolute_import, print_function, division
from six.moves import range
from six import integer_types

import numpy as np
import mdtraj as mdt
import os
import logging as log
from MDPlus import fastfitting, utils
import warnings

class Cofasu(object):

    """
    Construct a Cofau object from a list of Fasu objects.

    A Cofasu presents the data in one or many trajectory files in the form
    of a numpy array-like object.

    Args:

        fasulist (Fasu or list of Fasus): Define the trajectory data to
            be included.
        check (Optional, str): if check is None, fasus are considered 
            legitimate if they match in their second dimension (i.e., 
            number of atoms) If check is 'names', fasus must also match 
            atom names.  If check is 'masses', fasus must match for atom 
            masses, but not neccessarily names.
        comm (Optional, MPI communicator): If provided, trajectory I/O and
            processing is done in parallel.

    Attributes:

        avg (numpy array): mean coordinates as a [natoms, 3] array

    Other attributes are numpy array-like (see Examples)

    Examples::

        f1 = Fasu('protein.pdb','traj1.dcd', selection='name CA')
        f2 = Fasu('protein.pdb' 'traj2.xtc', selection='name CA')
        trajdata = Cofasu([f1, f2])
        print trajdata[34] # prints the 34th coordinate set
        print trajdata[:].mean(axis=0) # prints the average structure

    """
    def __init__(self, fasulist, check=None, comm=None):

        if comm is not None:
            rank = comm.Get_rank()
            size = comm.Get_size()
        else:
            rank = 0
            size = 1

        if not isinstance(fasulist, list):
            fasulist = [fasulist,]

        nfasus = len(fasulist)
        for ifas in range(nfasus):
            if fasulist[ifas].owner is None:
                fasulist[ifas].owner = size * ifas // nfasus

        for f in fasulist:
            if f.owner == rank:
                f._process()
            else:
                f.names = None
                f.masses = None
                f.shape = None
                f.x = None
                f.topology = None

        for f in fasulist:
            if comm is not None:
                f.masses = comm.bcast(f.masses, root=f.owner)
                f.names = comm.bcast(f.names, root=f.owner)
                f.shape = comm.bcast(f.shape, root=f.owner)
                f.topology = comm.bcast(f.topology, root=f.owner)
                if f.x is None:
                    f.x = np.array((1, f.shape[1], f.shape[2]), dtype='float32')
        
        nats = fasulist[0].shape[1]
        for i in range(1, len(fasulist)):
            if not fasulist[i].shape[1] == nats:
                raise ValueError('Fasus have mismatched number of atoms')

        if check is "names":
            nref = fasulist[0].names
            for i in range(1, len(fasulist)):
                if not (nref[:] == fasulist[i].names[:]).all:
                    raise ValueError('Fasus have mismatched atom names')

        elif check is "masses":
            mref = fasulist[0].masses
            for i in range(1, len(fasulist)):
                if not (mref[:] == fasulist[i].masses[:]).all:
                    raise ValueError('Fasus have mismatched atom masses')

        # Update fasu.frames slice definitions now file sizes are known:
        for i in range(len(fasulist)):
            if isinstance(fasulist[i].frames, slice):
                frames = fasulist[i].frames
                if frames.stop == None:
                    l = fasulist[i].shape[0]
                    start = frames.start
                    step = frames.step
                    stop = l
                    #stop = l + 1
                    if step is not None:
                        stop = stop * step
                    if start is not None:
                        stop += start
                    fasulist[i].frames = slice(start, stop, step) 

        self.masses = fasulist[0].masses # atom masses
        self.topology = fasulist[0].topology
        totframes = 0
        for f in fasulist:
            totframes += f.shape[0]
        # Shape of trajectory array (nframes, natoms, 3)
        self.shape = (totframes, fasulist[0].shape[1], fasulist[0].shape[2])
        self.fasulist = fasulist
        self.comm = comm
        self.avg = self.mean_structure() # mean structure

    def _x(self, key):
        """
        Private method to access coordinate data.
        """

        if isinstance(key, tuple):
            _key1 = key[0]
            if isinstance(_key1, integer_types):
                _key1 = slice(_key1, _key1+1)
            _key2 = key[1]
            if len(key) == 3:
                _key3 = key[2]
            else:
                _key3 = None

        elif isinstance(key, integer_types):
            _key1 = slice(key, key+1)
            _key2 = None
            _key3 = None

        elif isinstance(key, slice):
            _key1 = key
            _key2 = None
            _key3 = None

        totsize = self.shape[0]
        indx = list(range(totsize))[_key1]
        newsize = len(indx)

        reverse = False
        if indx[-1] < indx[0]:
            reverse = True
            indx = indx[::-1]
        maxval = 0
        j = 0
        x = []
        y = []
        sizes = [f.shape[0] for f in self.fasulist]
        for size in sizes:
            ix = []
            iy = []
            minval = maxval
            maxval += size
            while j < len(indx) and indx[j] < maxval:
                ix.append(j)
                iy.append(indx[j] - minval)
                j += 1
            x.append(ix)
            y.append(iy)
        if reverse:
            for ix in x:
                for v in ix:
                    v = len(indx) - v
            for iy in y:
                 iy = iy.reverse()

        xtmp = np.zeros((newsize, self.shape[1], self.shape[2]), dtype='float32')
        for i in range(len(self.fasulist)):
            f = self.fasulist[i]
            if self.comm is not None:
                
                rank = self.comm.Get_rank()
                if rank == f.owner:
                    xtmp[x[i]] = f.x[y[i]]
                xtmp[x[i]] = self.comm.bcast(xtmp[x[i]], root=f.owner)
            else:
                xtmp[x[i]] = f.x[y[i]]
        if _key2 is not None:
            if _key3 is not None:
                xtmp = xtmp[(slice(0,len(xtmp)), _key2, _key3)]
            else:
                xtmp = xtmp[(slice(0,len(xtmp)), _key2)]
        if len(xtmp) == 1:
            xtmp = xtmp[0]
        return xtmp

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, key):
        return self._x(key)

    def mean_structure(self):
        """
        Return the mean structure, without any fitting process.

        Returns:
            crds ([natoms, 3] numpy array): coordinates of mean structure

        """
        if self.comm is None:
            rank = 0
        else:
            rank = self.comm.Get_rank()

        sum = np.zeros((self.shape[1], self.shape[2]), dtype='float32')
        for f in self.fasulist:
            if f.owner == rank:
                sum += f.x.sum(axis=0)
        if self.comm is not None:
            tsum = self.comm.allreduce(sum)
        else:
            tsum = sum

        return  tsum/len(self)

    def align(self, target=None, weighted=False, procrustes=False,
    error=0.0001, maxcyc=10):
        """
        Align the frames in a trajectory to some reference structure.
        
        Least-squares fits each frame to some reference structure, with
        optional mass-weighting.

        Args:
            target (Optional, [N,3] numpy array): A reference structure to fit 
                to; if not provided, the first frame in the first trajectory is
                used.

            weighted (Optional, bool, default=False): If specified, mass-
                weighted fitting is done.

            procrustes (Optional, bool, default=False): If True, procrustes 
                iterative fitting is done to convergence.

            error (Optional, float, default=0.0001): Defines the target error 
                for the procrustes fit.

            maxcyc (Optional, int, default=10): Defines the maximum number of 
                iterations for the procrustes method.

        Examples::

            c = Cofasu([fasu1, fasu2])
            c.align() # align to 1st frame
            c.align(target=c[34], procrustes=True) # iterativly align to 34th frame
            c.align(weighted=True) # Align with mass-weighting

        """

        if target is None:
            targ = self[0]
        else:
            targ = target

        if weighted:
            weights = self.masses
        else:
            weights = np.ones_like(self.masses)
        weights = np.array([weights,] * 3).T

        if self.comm is None:
            rank = 0
        else:
            rank = self.comm.Get_rank()
        for f in self.fasulist:
            if f.owner ==  rank:
                f.x = fastfitting.fitted_traj(f.x, targ, weights)

        self.avg = self.mean_structure()
            
        if not procrustes:
            return

        err = self.avg - targ
        err = (err*err).mean()
        cycle = 1
        while err > error and cycle < maxcyc:
            target = self.avg
            oldavg = self.avg
            for f in self.fasulist:
                if f.owner ==  rank:
                    f.x = fastfitting.fitted_traj(f.x, target, weights)

            self.avg = self.mean_structure()
            
            err = self.avg - oldavg
            err = (err*err).mean()
            cycle += 1
        if self.comm is None or rank == 0:
            log.debug('Procrustes converged in {} cycles'.format(cycle)
                      + ' with error {}'.format(err))

    def write(self, filename, coordinates=None):
        """
        Write selected data to an output trajectory file.

        Args:
            filename (str): Name of the file to be written. All MDTraj-
                supported formats are available, selected by the filename
                extension.

            coordinates (Optional, [nframes, natoms, 3] numpy array):
                The data to be written, else all frames in the Cofasu will be 
                output.

        Examples::

            trajdata.write('output.dcd') # Outputs all frames
            start = trajdata[:20]
            trajdata.write('twentyframes.xtc', start) # Just 20 frames.
        """

        # Note: currently ignores box data.
        ext = os.path.splitext(filename)[1].lower()
        needs_topology = ext in ['.gro', '.pdb']

        if self.comm is not None:
            rank = self.comm.Get_rank()
        else:
            rank = 0


        if coordinates is None:
            coordinates = self[:]

        if ext in ['.xtc', '.trr']:
            coordinates = coordinates * 0.1
        if ext in '.gro' and coordinates.ndim == 2:
            coordinates = np.array([coordinates,])
        if rank == 0:
            top = self.topology
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                with mdt.open(filename, 'w') as f:
                    if needs_topology:
                        f.write(coordinates, top)
                    else:
                        f.write(coordinates)

class Fasu(object):
    """
    An object that defines the data to be extracted from trajectory files.

    A Fasu object is a wrapper round a trajectory file that defines which
    frames and atoms are to be extracted for analysis, and if certain
    coordinate transformations are required (centering and/or imaging).

    Args:
        trajectory (str or [str] or Trajectory): Name(s) of trajectory 
            file(s), or an mdtraj.Trajectory.
        topology (Optional, str): Name of a topology file compliant with the 
            trajectory files(s). Will be required if the trajectory file
            does not provide topology information itself.
        frames (Optional, slice object or numpy array): Selection of 
            trajectory frames to include.
        selection (Optional, str): mdtraj-compliant atom selection
            string.
        centre (Optional, str): An mdtraj-compliant atom selection string 
            that defines atoms whose geometric centre will be moved to the 
            centre of the periodic box (if there is one), or to the origin 
            (if there isn't).
        pack_into_box (Optional, bool): if True, all coordinates will be 
            imaged into the primary unit cell, after any centering has been
            done.

    Examples::

        f = Fasu(trajectory.dcd', top='protein.pdb')
        f = Fasu[['traj1.dcd', 'traj2.dcd'], top='protein.pdb')
        f = Fasu('trajectory.xtc', top='protein.pdb', frames=slice(0,50,5))
        f = Fasu('trajectory.nc', top='protein.pdb', selection='name CA')
        f = Fasu('traj.dcd', top='prot.pdb', centre='(resid 20) and (name CA)',
                 pack_into_box=True)

    """

    def __init__(self, trajectory, top=None, frames=None, selection='all', 
    centre=None, pack_into_box=False):

        if top is not None:
            test = open(top, 'r')
            test.close()
        self.topfile = top # name of topology file

        if not isinstance(trajectory, mdt.Trajectory):
            self.trajectory = None
            if not isinstance(trajectory, list):
                self.trajfiles = [trajectory,]
            else:
                self.trajfiles = trajectory

            for t in self.trajfiles:
                test = open(t, 'r')
                test.close()
        else:
            self.trajectory = trajectory
            self.trajfiles = None

        if frames is not None:
            if not (isinstance(frames, slice) or 
                    isinstance(frames, np.ndarray)):
                raise TypeError('frames must be a slice object or numpy array')
        if frames is None:
            frames = slice(0, None, 1)

        self.frames = frames # number of frames
        self.selection = selection # MDTRaj selection string
        self.centre = centre
        self.pack_into_box = pack_into_box
        self.owner = None
        self.processed = False

    def _process(self):
        """
        Private function that processes the Fasu definition

        """
        if self.processed:
            return

        if self.trajectory is None:
            ext = os.path.splitext(self.trajfiles[0])[1].lower()
            if not ext in ['.gro', '.pdb']:
                self.trajectory = mdt.load(self.trajfiles, top=self.topfile)
            else:
                self.trajectory = mdt.load(self.trajfiles)

        self.sel = self.trajectory.top.select(self.selection)
        self.trajectory.top = self.trajectory.top.subset(self.sel)
        self.topology = self.trajectory.top
        self.trajectory.xyz = self.trajectory.xyz[:, self.sel, :]
        if self.trajectory.n_atoms == 0:
            raise ValueError('Atom selection matches no atoms.')

        masses = [atom.element.mass for atom in self.trajectory.top.atoms]
        masses = np.array(masses, dtype='float32')
        names = [self.trajectory.top.atom(i).name for i in
                 range(self.trajectory.n_atoms)]

        if self.frames is not None:
            x = np.array(self.trajectory.xyz[self.frames], dtype='float32', 
                         order='F')
        else:
            x = np.array(self.trajectory.xyz, dtype='float32', order='F')

        if self.centre is not None:
            c = self.trajectory.top.select(self.centre)
            if len(c) == 0:
                raise ValueError('Atom selection for centering matches no atoms.')
            for i in range(len(x)):
                try:
                    cx = x[i][c].mean(axis=0)
                except IndexError:
                    print(c)
                    raise
                if self.trajectory.unitcell_vectors is None:
                    shift = -cx
                else:
                    shift = self.trajectory.unitcell_vectors[i].diagonal()/2 - cx
                x[i] = x[i] + shift

        if self.pack_into_box:
            x = fastfitting.pib(x, self.trajectory.unitcell_vectors)

        self.x = x * 10.0 # convert from nanometres to angstroms.
        self.masses = masses
        self.names = names
        self.trajectory = self.trajectory[0] # truncate to save memory
        self.shape = x.shape
        self.processed = True



from __future__ import absolute_import, print_function, division
from six.moves import range

import numpy as np
import pandas as pd
from scipy.linalg import eigh
from scipy.optimize import minimize_scalar
import struct
import logging as log
import sys
from time import time
import warnings
import mdtraj as mdt
from MDPlus.core import Fasu, Cofasu
from MDPlus import fastfitting
import networkx as nx

def load(filename):
    """
    Initialises a new Pcamodel object with the data from the given file.

    Args:
        filename (str): Name of the .pcz format file to be loaded.

    """
    try:
        filehandle = open(filename, 'rb')
    except IOError as e:
        print("Problems while trying to open {}.".format(filename))
        print("I/O error({0}): {1}".format(e.errno, e.strerror))
        raise
    data = filehandle.read(104)
    try:
        filetype = data[0:4].decode('utf-8')
    except:
        filetype = None

    if not filetype in ['PCZ1']:
        print("Error: {}: unrecognised pcz file format".format(filename))
        raise IOError("Error: unrecognised pcz file format")

    if filetype == 'PCZ1':
        keydata = struct.unpack('4s80s4if', data)
        version = keydata[0].decode('utf-8')
        title = keydata[1].decode('utf-8')
        n_atoms = keydata[2]
        nbonds = keydata[3]
        n_frames = keydata[4]
        n_vecs = keydata[5]
        totvar = keydata[6]

        atomnames = []
        resnames = []
        chainnames = []
        bonds = []
        elements = []

        data = filehandle.read(4 * n_atoms)
        atomseq = np.array(struct.unpack('{}i'.format(n_atoms),
                           data))
        for i in range(n_atoms):
            data = filehandle.read(4) 
            aname = struct.unpack('4s', data)
            atomnames.append(str(aname[0].decode('utf-8')))
            n = str(aname[0]).strip().upper()
            e = n[0]
            if e.isdigit():
                e = n[1]
            elements.append(e)
        for i in range(n_atoms):
            data = filehandle.read(3) 
            rname = struct.unpack('3s', data)
            resnames.append(rname[0].decode('utf-8'))
        data = filehandle.read( 4 * n_atoms)
        resseq = np.array(struct.unpack('{}i'.format(n_atoms),
                          data))
        for i in range(n_atoms):
            data = filehandle.read(1) 
            cname = struct.unpack('1s', data)
            chainnames.append(cname[0].decode('utf-8'))

        data = filehandle.read( 4 * nbonds)
        bindx = struct.unpack('{}h'.format(2 * nbonds), data)
        bindx = np.array(bindx).reshape((-1, 2))

        data = filehandle.read(12 * n_atoms) 
        acme = struct.unpack('{}f'.format(3 * n_atoms), data)
        acme = np.array(acme).reshape((-1, 3))
        data = filehandle.read(12 * n_atoms) 
        avg = struct.unpack('{}f'.format(3 * n_atoms), data)
        avg = np.array(avg).reshape((-1, 3))
        data = [(atomseq[i], atomnames[i], elements[i], resseq[i], resnames[i], chainnames[i], "") for i in range(n_atoms)]
        atoms = pd.DataFrame(data, columns=["serial", "name", "element", "resSeq", "resName", "chainID", "segmentID"])
        topology = mdt.Topology.from_dataframe(atoms, bindx)

        data = filehandle.read(4 * n_vecs)
        evals = struct.unpack('{}f'.format(n_vecs), data)
        evals = np.array(evals)

        evecs = np.zeros((n_vecs, 3 * n_atoms))
        for i in range(n_vecs):
            data = filehandle.read(12 * n_atoms)
            evecs[i] = struct.unpack('{}f'.format(3 * n_atoms), data)

        data = filehandle.read(4 * n_vecs)
        p0 = struct.unpack('{}f'.format(n_vecs), data)
        p0 = np.array(p0)
            
        data = filehandle.read(4 * n_vecs)
        pinc = struct.unpack('{}f'.format(n_vecs), data)
        pinc = np.array(pinc)

        projs = np.zeros((n_vecs, n_frames))
        prj = np.zeros((n_vecs))
        for i in range(n_frames):
            data = filehandle.read(2 * n_vecs)
            prj[:] = struct.unpack('{}h'.format(n_vecs), data)
            projs[:,i] = (prj * pinc) + p0

        regready = False
    else:
        raise IOError('Error: unrecognised pcz file format')

    return Pcamodel(evecs, evals, projs, avg, acme, topology, totvar,
                    version=version, title=title)

def fromtrajectory(traj, quality=90.0, req_evecs=None, fastmethod=False,
                   weighted=False):
    """
    Initialises a new Pcamodel object with the data from the given
    mdtraj-like object.

    Arguments:
        traj: An mdtraj trajectory, or similar.
        quality (Opt, float): Percentage of variance to be explained.
            Defaults to 90.0 (90%).
        req_evecs (Optional, int): Number of eigenvectors to calculate. If
            set, overrides quality setting.
        fastmethod (optional, bool): If True, an approximate fast matrix
            diagonalizaion method is used.
        weighted (optional, bool): If True, PCA is done on the mass-weighted
            covariance matrix.

    >>> topfile = "../../../test/2ozq.pdb"
    >>> trjfile = "../../../test/2ozq.dcd"
    >>> from MDPlus.core import Fasu, Cofasu
    >>> t = Cofasu(Fasu(trjfile, top=topfile, 
    ...            selection='name CA and resid 1 to 10'))
    >>> p = fromtrajectory(t)

    
    The quality setting defaults to 90%:
    >>> p = fromtrajectory(t, quality=95)
    >>> p = fromtrajectory(t, quality=120)
    Traceback (most recent call last):
    ...
    ValueError: Pcz: quality must lie in the range 0:100.

    If fastmethod=True then a fast approximate diagonalisation
    method is used.

    >>> f = Fasu(trjfile, top=topfile, selection='resid 1 to 30')
    >>> t = Cofasu(f)
    >>> p1 = fromtrajectory(t)
    >>> ev1 = p1.evals
    >>> p2 = fromtrajectory(t, fastmethod=True)
    >>> ev2 = p2.evals
    >>> print(np.allclose(ev1, ev2))
    True

    """
    if isinstance(traj, mdt.Trajectory):
        traj = Cofasu(Fasu(traj))
    n_atoms = traj.shape[1]
    n_frames = traj.shape[0]
    acme = traj[0]
    comm = traj.comm
    if comm is not None:
        rank = comm.Get_rank()
        size = comm.Get_size()
    else:
        rank = 0
        size = 1
        
    if quality < 0 or quality > 100:
        raise ValueError('Pcz: quality must lie in the range 0:100.')

    if rank == 0:
        log.debug('Pcz: {0} atoms and {1} snapshots'.format(n_atoms, 
                 n_frames))
    if rank == 0:
        log.info('Pcz: least-squares fitting snapshots...')
    time_avg_0 = time()
    traj.align(target=acme, procrustes=True, weighted=weighted)
    avg = traj.avg
    time_avg_1 = time()
    if rank == 0:
        log.debug( 'Pcz: Time for trajectory fitting: '
                 + '{0:.2f} s'.format(time_avg_1 - time_avg_0))

    if rank == 0:
         log.info('Pcz: calculating covariance matrix...')
    time_cov_0 = time()
    if n_frames < (3 * n_atoms):
        fastmethod = True
    if fastmethod:
        # adapted from Ian Dryden's R code. If you have
        # n atoms and p snapshots, then the conventional
        # way to do the pca is to calculate the [3n,3n]
        # covariance matrix and then diagonalise that.
        # However if p < 3n, then the last 3n-p eigenvectors
        # and values are meaningless anyway, and instead
        # you can calculate the [p,p] covariance matrix,
        # diagonalise that, and then do some
        # data massaging to recover the full eigenvectors
        # and eigenvalues from that. Here we extend the
        # approach to situations where 3n is just too big
        # for diagonalisation in a reasonable amount of
        # time, by just taking a selection of snapshots
        # from the full set (<< 3n) and applying this
        # approach. Obviously this is an approximate
        # method, but it may be good enough.
        if rank == 0:
            log.info("Using fast approximate diagonalisation method")
        nsamples = min(1000,n_frames)
        stepsize = n_frames//nsamples
        tmptrj = traj[::stepsize]
        tmptrj = tmptrj - avg
        tmptrj = tmptrj.reshape((-1, 3 * n_atoms))

        cv = np.dot(tmptrj, tmptrj.T)/nsamples
    else:
        n3 = n_atoms * 3
        lcv = np.zeros((n3, n3))
        nlocal = 0
        for f in traj.fasulist:
            if f.owner == rank:
                dx = (f.x - avg).reshape((-1, n3))
                lcv += np.dot(dx.T, dx.conj())
                nlocal += 1
        log.debug('Process {} owns {} fasus.'.format(rank, nlocal))
        if comm is None:
            cv = lcv
        else:
            comm.Barrier()
            cv = np.zeros((n3, n3))
            comm.Allreduce(lcv, cv)
        cv = cv / n_frames
        if rank == 0:
            log.debug("Global CM calculated")

    if weighted:
        m = np.sqrt(np.vstack((traj.masses, traj.masses, traj.masses)).T.flatten())
        m2 = np.dot(m.T, m)
        cv = cv * m2
        
    time_cov_1 = time()
    if rank == 0:
        log.debug( 'Pcz: Time for calculating covariance matrix: '
                 + '{0:.2f} s\n'.format(time_cov_1 - time_cov_0))
        log.debug('Total variance = {}'.format(cv.trace()))
    if rank == 0:
        if weighted:
            log.info('Pcz: diagonalizing mass-weighted covariance matrix...')
        else:
            log.info('Pcz: diagonalizing covariance matrix...')
    time_diag_cov_0 = time()

    if fastmethod:
        w, v = eigh(cv)
        vv = np.zeros(nsamples)
        z = np.dot(tmptrj.T,v)
        for i in range(nsamples):
            vv[i] = np.sqrt((z[:,i]*z[:,i]).sum())
            z[:,i] = z[:,i]/vv[i]

            w2 = np.sqrt(abs(w/nsamples))*vv
        w = w2[w2.argsort()]
        v = z[:,w2.argsort()]
        cs = np.cumsum(w[::-1])
        totvar = cs[-1]
        tval = cs[-1] * quality / 100
        i = 0
        while cs[i] < tval:
            i += 1
        n_vecs = i + 1
    else:
        if rank == 0:
            w, v = eigh(cv)
        else:
            w = None
            v = None
        if comm is not None:
            w = comm.bcast(w, root=0)
            v = comm.bcast(v, root=0)
        cs = np.cumsum(w[::-1])
        totvar = cs[-1]
        tval = cs[-1] * quality / 100
        i = 0
        while cs[i] < tval:
            i += 1
        n_vecs = i + 1
        if req_evecs is not None:
            n_vecs = req_evecs
            if req_evecs > len(cv) - 1:
                raise ValueError('Pcz: you asked for {0} eigenvectors but there'
                              + ' are only {1} available.'.format(req_evecs, 
                                                                  len(w)))
    evals = w[-1:-(n_vecs + 1):-1]
    evecs = v[:, -1:-(n_vecs + 1):-1].T

    time_diag_cov_1 = time()
    if rank == 0:
        log.debug( 'Pcz: Time for diagonalizing covariance matrix: '
                 + '{0:.2f} s\n'.format(time_diag_cov_1 - time_diag_cov_0))

    time_proj_calc_0 = time()
    if rank == 0:
        log.info('Pcz: calculating projections...')

    plist = []
    n3 = 3 * n_atoms
    for f in traj.fasulist:
        if f.owner == rank:
            dx = (f.x - avg).reshape((-1, n3))
            plist.append(np.dot(dx, evecs.T).T)

    if len(plist) == 0:
        lproj = np.zeros((0, 0))
    else:
        lproj = np.hstack(plist)
    if comm is not None:
        gproj = comm.allgather(lproj)
        projs = np.hstack([p for p in gproj if p.size > 0])
    else:
        projs = lproj
        
    time_proj_calc_1 = time()
    if rank == 0:
        log.debug('Pcz: Time for calculating projections: '
                 + '{0:.2f} s'.format(time_proj_calc_1 - time_proj_calc_0))
    
    return Pcamodel(evecs, evals, projs, avg, acme, traj.topology, totvar)


class Pcamodel(object):
    """
    PCA analysis of trajectory data.

    Defines the attributes and methods of an object that describes a PCA model
    built from (probably) MD trajectory data.

    Attributes:
        n_atoms (int): number of atoms
        n_frames (int): number of frames
        n_vecs (int): Number of eigenvectors.
        title: (str): Title (usually taken from trajectory file).
        refxyz ([n_atoms, 3] numpy array): Mean structure.
        acme ([n_atoms, 3] numpy array): Coordinates of a structure with 
            "good" geometry - typically the first frame of the original 
            trajectory.
        topology: The MDTraj topology
        totvar (float): The total variance in the original trajectory.
        evals ([n_vecs] numpy array): Eigenvalues, in descending order of size.
        evecs ([n_vecs, 3*n_atoms] numpy array): Eigenvectors, in same order as
            eigenvectors.
        projs ([n_vecs, n_frames] numpy array): Projection data.

    Methods:

    """

    def __init__(self, evecs, evals, projs, refxyz, acme, topology, totvar, 
                 title=None, version='PCZ1'):
        """
        Initialises a new Pcamodel object.

        """
        self.evecs = evecs
        self.evals = evals
        self.projs = projs
        self.refxyz = refxyz
        self.acme = acme
        self.topology = topology
        self.totvar = totvar
        self.rank = 0
        self.n_atoms = self.topology.n_atoms
        self.n_frames = self.projs.shape[1]
        self.n_vecs = len(evals)
        self.title = title
        if self.title is None:
            self.title = 'Created by pyPcazip'
        
        self.version = version
        self._bondinit()


    def scores(self, framenumber):
        """
        Return the scores (projections) corresponding to a chosen snapshot.

        Args:
            framenumber (int): index of trajectory frame

        Returns:
            scores ([n_vecs] numpy array): Scores for each eigenvector.

        Example:

        >>> topfile = "../../../test/2ozq.pdb"
        >>> trjfile = "../../../test/2ozq.dcd"
        >>> from MDPlus.core import Fasu, Cofasu
        >>> t = Cofasu(Fasu(trjfile, top=topfile, selection='name CA'))
        >>> p = fromtrajectory(t)
        >>> s = abs(p.scores(12))
        >>> print(np.allclose(s[3], 0.5800652, rtol=0.001, atol=0.001))
        True

        """
        return self.projs.T[framenumber]


    def frame(self,framenumber, regularize=True):
        """
        Method to return the coordinates of the given frame.
        
        Args:
            framenumber (int): index of selected frame
            regularize (bool): if True, the geometry
                (bond lengths and angles are adjusted to match those in the
                "acme" structure as far as possible, without changing the
                PC scores of this frame).

        Retuns:
            crds ([natom, 3] numpy array): coordinates


        Example:

        >>> topfile = "../../../test/2ozq.pdb"
        >>> trjfile = "../../../test/2ozq.dcd"
        >>> from MDPlus.core import Fasu, Cofasu
        >>> t = Cofasu(Fasu(trjfile, top=topfile, selection='name CA'))
        >>> t.align(t[0])
        >>> p = fromtrajectory(t, quality=95)
        >>> ref = t[5]
        >>> x = p.frame(5)
        >>> print( (abs(x - ref)).mean() < 0.19)
        True

        """
        if(framenumber >= self.n_frames):
            return None
        else:
            scores = self.scores(framenumber)
            return self.unmap(scores, regularize=regularize)


    def closest(self, scores):
        """
        Returns the index of the frame with scores closest to the target values.

        Args:
            scores (numpy array): target scores. If the scores vector has less
                than nvec elements, the least significant are assumed to be zero.

        Returns:
            indx (int): The index of the snapshot whose scores are closest to
                the input set.
        """
        ns = len(scores)
        temp = self.projs

        best = 0
        err = ((temp[0:ns, 0] - scores) * (temp[0:ns, 0] - scores)).sum()
        for frame in range(self.n_frames):
            newerr = ((temp[0:ns, frame] - scores) 
                      * (temp[0:ns, frame] - scores)).sum()
            if newerr < err:
                err = newerr
                best = frame
        return best

    def _bondinit(self):
        """ Private function to set up bond length and angle data 

        Initialises the bond and angle (1,3 atom pairs) arrays used for
        the regularizaion process.
        """

        bonds = [bond for bond in self.topology.bonds]
        bindx = []
        for bond in bonds:
            bindx.append((bond[0].index, bond[1].index))
        i = 0
        self.nbonds = len(bindx)
        # check for complete connectivity:
        G = nx.Graph()
        G.add_edges_from(bindx)
        cc = list(nx.connected_components(G))
        if len(cc) > 1 or self.nbonds == 0:
            log.debug('Molecule is fragmented - trying to fix...')
            bindx = self._find_bonds(self.acme)
            self.nbonds = len(bindx)
            G.clear()
            G.add_edges_from(bindx)
            cc = list(nx.connected_components(G))
            if len(cc) > 1 or self.nbonds == 0:
                log.debug('Cannot fix - no regularization possible')
                self.fragmented = True
            else:
                self.fragmented = False
                log.debug('Found {} bonds.'.format(self.nbonds))
        else:
            self.fragmented = False
            
        if not self.fragmented:
            b13 = []
            for b0 in bindx[:-1]:
                i += 1
                for b1 in bindx[i:]:
                    if b0[0] == b1[0]:
                        b13.append((b0[1], b1[1]))
                    if b0[1] == b1[0]:
                        b13.append((b0[0], b1[1]))
            self.b13 = bindx + b13
            self.lref = self._bond_lengths(self.acme, self.b13)
        self.bindx = bindx

    def _find_bonds(self, x):
        """
        Find likely bonds, based on stored topology and coordinates in x
        """
        ref = mdt.Trajectory(x, self.topology)
        pairs = []
        t = self.topology
        for i in range(ref.n_atoms - 1):
            for j in range(i + 1, ref.n_atoms):
                pairs.append((i, j))
        d = mdt.compute_distances(ref, pairs, periodic=False)[0]
        r = [(t.atom(i).element.radius + t.atom(j).element.radius) for i,j in pairs]
        d = np.array(d)
        r = np.array(r) * 0.53
        bindx = []
        for i in range(len(pairs)):
            if d[i] - r[i] < 0:
                bindx.append(pairs[i])
        return bindx
        

    def _bond_lengths(self, x, bondlist):
        """
        calculate lengths of bonds in x
        """
        llist = []
        for b in bondlist:
            dx = x[b[0]] - x[b[1]]
            llist.append(np.sqrt((dx * dx).sum()))
        return np.array(llist)

    def _lerror(self, x, bondlist, lref):
        """
        bond length error metric
        """
        assert len(bondlist) == len(lref)
        lx = self._bond_lengths(x, bondlist)
        dl = lx - lref
        return (dl * dl).mean()

    def _bond_shake(self, x, bondlist, lref, evecs, tol=0.005):
        """
        SHAKE type process to regularize bond lengths and angles

        Input coordinates x are refined to minimize error
        in the bond and 1-3 atom distances relative to those in the "acme"
        structure. The coordinate shifts are restricted to dimensions not in
        the essential subspace. 
        """
        if len(bondlist) == 0:
            return x
        if self.fragmented:
            return x

        xshift = np.zeros_like(x)
        errinit = self._lerror(x, bondlist, lref)
        errold = errinit + 2.0 * tol
        errnew = errinit
        def fun(sfac, x, xshift, bondlist, lref):
            return self._lerror(x + xshift * sfac, bondlist, lref)
        while errold - errnew > tol:
            xshift[:] = 0.0
            errold = errnew
            for b in range(len(bondlist)):
                i = bondlist[b][0]
                j = bondlist[b][1]
                dx = x[i] - x[j]
                l = np.sqrt((dx * dx).sum())
                dl = lref[b] - l
                dx = 0.5 * (dx * dl / l)
                xshift[i] += dx
                xshift[j] -= dx

            xshift = xshift.flatten()
            for e in evecs:
                xshift -= e * np.dot(e, xshift)
            xshift = xshift.reshape((-1, 3))
            result = minimize_scalar(fun, bounds=(0.0, 1.0), args=(x,
                                 xshift, self.b13, self.lref))
            sfac = result.x
            errnew = result.fun
            x = x + xshift * sfac
        return x

    def unmap(self,scores, regularize=True, tol=0.005):
        """
        Return the coordinates corresponding to a given set of scores. 

        Args:
            scores (numpy array): a list of scores. If scores has less than
                nvec elements, the least significant are assumed to be zero.
            regularize (bool): if True, coordinates are regularised - bond 
                lengths and angles are refined to match as far as possible
                those in self.acme.


        Returns:
            crds ([natom, 3] numpy array): coordinates.

        Example:

        >>> topfile = "../../../test/2ozq.pdb"
        >>> trjfile = "../../../test/2ozq.dcd"
        >>> from MDPlus.core import Fasu, Cofasu
        >>> t = Cofasu(Fasu(trjfile, top=topfile, selection='name CA'))
        >>> p = fromtrajectory(t)
        >>> a = p.refxyz
        >>> a2 = p.unmap(np.zeros(p.n_vecs))
        >>> print((abs(a - a2)).mean() < 0.001)
        True

        """
        x = self.refxyz
        for i in range(self.n_vecs):
            x = x + (self.evecs[i]*scores[i]).reshape((self.n_atoms,3))

        if regularize and (self.fragmented == False):
            xreg = self._bond_shake(x, self.b13, self.lref, self.evecs, tol)
            x = xreg
            
        return x

    def map(self,crds):
        """
        Method to map an arbitrary coordinate set onto the PC model. 
        
        The coordinate set should be a (natom,3) array-like object that matches
        (for size) what's in the pczfile. An array of scores will be 
        returned, one value for each eignevector in the pcz file.

        Args:
            crds ([natom, 3] numpy array): coordinates.

        Returns:
            scores ([n_vecs] numpy array): list of scores.

        Example:

        >>> topfile = "../../../test/2ozq.pdb"
        >>> trjfile = "../../../test/2ozq.dcd"
        >>> from MDPlus.core import Fasu, Cofasu
        >>> t = Cofasu(Fasu(trjfile, top=topfile, selection='name CA'))
        >>> p = fromtrajectory(t)
        >>> m = p.scores(10)
        >>> crds = t[10]
        >>> print(np.allclose(abs(p.map(crds)),abs(m), rtol=0.001, atol=0.001))
        True

        """
        c = fastfitting.fitted(crds, self.refxyz)
        c = c - self.refxyz
        prj = np.zeros(self.n_vecs)
        for i in range(self.n_vecs):
             prj[i]=(np.dot(c.flatten(),self.evecs[i]))
        return prj
  

    def write(self, filename,  version='PCZ1', title='Created by pcz.write()'):
        """
        Write out the PCZ file.

        Args:
            filename (str): Name of .pcz file.
            version (str): Pcz version (format) - only 'PCZ1' supported here.
            title (str): Title for .pcz file.

        """

        self.version = version # Pcz file version (format)
        if self.rank != 0:
            return


        if self.version == 'PCZ1':
            log.debug("Using "+self.version+" format")
            f = open(filename, 'wb')
            f.write(struct.pack('4s', self.version.encode('utf-8')))
            f.write(struct.pack('80s', title.encode('utf-8')))
            f.write(struct.pack('4if', 
                                self.n_atoms, self.nbonds, self.n_frames, 
                                self.n_vecs, self.totvar))

            chainnames = [chr(ord('A') + i) for i in range(26)]
            top = self.topology
            atoms = top.atoms
            for a in atoms:
                f.write(struct.pack('i', a.serial))
            for a in range(self.n_atoms):
                aname = top.atom(a).name
                if len(aname) == 1:
                    aname = ' ' + aname + '  '
                elif len(aname) == 2:
                    aname = ' ' + aname + ' '
                elif len(aname) == 3:
                    aname = ' ' + aname
                f.write(struct.pack('4s', aname.encode('utf-8')))
            for a in range(self.n_atoms):
                rname = top.atom(a).residue.name
                if len(rname) == 1:
                    rname = rname + '  '
                elif len(rname) == 2:
                    rname = rname + ' '
                f.write(struct.pack('3s', rname.encode('utf-8')))
            for a in range(self.n_atoms):
                f.write(struct.pack('i', top.atom(a).residue.resSeq))
            for a in range(self.n_atoms):
                i = top.atom(a).residue.chain.index % 26
                f.write(struct.pack('1s', chainnames[i].encode('utf-8')))
            for bond in self.bindx:
                bond = np.array(bond).astype(np.int16)
                f.write(struct.pack('2h', bond[0], bond[1]))
            for v in self.acme.flatten():
                f.write(struct.pack('f', v))
                

            for v in self.refxyz.flatten():
                f.write(struct.pack('f', v))
            for i in range(self.n_vecs):
                f.write(struct.pack('f', self.evals[i]))
            for i in range(self.n_vecs):
                for v in self.evecs[i]:
                    f.write(struct.pack('f', v))

            p0 = (self.projs.max(axis=1) + self.projs.min(axis=1)) / 2
            for v in p0:
                f.write(struct.pack('f', v))

            pinc = (self.projs.max(axis=1) - self.projs.min(axis=1)) / 65534
            for v in pinc:
                f.write(struct.pack('f', v))

            for i in range(self.n_frames):
                prj = ((self.projs[:, i] - p0) / pinc).astype(np.int16)
                for ipr in prj: 
                    f.write(struct.pack('h', ipr))
            f.close()
            return
          
        else:
            raise TypeError('Only PCZ1 format supported')


    def maha(self, scores):
        """
        Calculate the Mahalanobis from a set of score deltas.

        Args:
            scores (numpy array): list of scores  - typically differences
                between the scores for two configurations. If fewer than n_vecs
                elements, the least sigificant are presumed to be zero.

        Returns:
            mdist (float): the Mahalanobis distance.
        """
        mdist = 0.0
        s = scores[:self.n_vecs]
        for i in range(len(s)):
            mdist += s[i] * s[i] / self.evals[i]

        return np.sqrt(mdist)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""  Setup script. Used by easy_install and pip. """

from __future__ import absolute_import, print_function, division

import os
import sys
import re
import textwrap


"""Some functions for checking and showing errors and warnings."""
def _print_admonition(kind, head, body):
    tw = textwrap.TextWrapper(initial_indent='   ', subsequent_indent='   ')

    print(".. {0}:: {1}".format(kind.upper(), head))
    for line in tw.wrap(body):
        print(line)


def exit_with_error(head, body=''):
    _print_admonition('error', head, body)
    sys.exit(1)


def print_warning(head, body=''):
    _print_admonition('warning', head, body)


def check_import(pkgname, pkgver):
    """ Check for required Python packages. """
    try:
        mod = __import__(pkgname)
        if mod.__version__ < pkgver:
            raise ImportError
    except ImportError:
        exit_with_error("Can't find a local {0} installation with version >= {1}. "
                        "Pypcazip needs {0} {1} or greater to compile and run! "
                        "Please read carefully the ``README`` file.".format(pkgname, pkgver))

    print("* Found {0} {1} package installed.".format(pkgname, mod.__version__))
    globals()[pkgname] = mod


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


"""Discover the package version"""
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
VERSIONFILE = "pypcazip/pcazip/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RunTimeError("Unable to find version string in {}.".format(VERSIONFILE))


"""Check Python version"""
print("* Checking Python version...")
if sys.version_info[:2] < (2, 7) or (3, 0) <= sys.version_info[0:2] < (3, 4):
    exit_with_error("You need Python 2.7 or 3.4+ to install pyPcazip!")
print("* Python version OK!")


"""Check minimum versions of NumPy, SciPy, Cython required."""
min_numpy_version = '1.0.3'
min_scipy_version = '0.10.0'
min_Cython_version = '0.19.0'

check_import('numpy', min_numpy_version)
check_import('scipy', min_scipy_version)
check_import('Cython', min_Cython_version)


"""Set up pyPcazip."""
from setuptools import setup, find_packages
from setuptools import Extension as Ext
class Extension(Ext, object):
    pass
from Cython.Build import cythonize

setup_args = {
    'name':             "pyPcazip",
    'version':          verstr,
    'description':      "PCA-based molecular dynamics trajectory file compression and analysis.",
    'long_description': "PCA-based molecular dynamics trajectory file compression and analysis.",
    'author':           "The University of Nottingham & BSC",
    'author_email':     "charles.laughton@nottingham.ac.uk",
    'url':              "https://bitbucket.org/ramonbsc/pypcazip/overview",
    'download_url':     "https://bitbucket.org/ramonbsc/pypcazip/get/{}.tar.gz".format(verstr),
    'license':          "BSD license.",

    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Environment :: Console',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],

    'packages': find_packages('pypcazip'),
    'package_dir': {'': 'pypcazip'},

    'scripts': ['scripts/pyPcazip',
                'scripts/pyPcaunzip',
                'scripts/pyPczcomp',
                'scripts/pyPczdump',
                'scripts/pyPczplot',
                'scripts/pyPczclust'],

    'install_requires': ['numpy',
                         'scipy',
                         'cython',
                         'mdtraj',
                         'argparse',
                         'six',
                         'networkx',
                         'pandas'],

    'zip_safe': False,

    'ext_modules':  cythonize(Extension(
                        name='MDPlus.fastfitting',
                        sources=['pypcazip/MDPlus/fastfitting/fastfitting.pyx'],
                        include_dirs=[numpy.get_include()]))
}

setup(**setup_args)



#!/usr/bin/env python

from __future__ import absolute_import, print_function, division

import sys
import logging as log

import numpy as np

from MDPlus.analysis import pca
import mdtraj as mdt
from mdtraj.geometry import alignment

def pczcomp(args):
    """
    Pczcomp compares two .pcz files. It reports on the RMSD between the
    two average structures, the dot product matrix, the subspace overlap,
    and the average maximum dot product. In addition (unless the '--quick'
    option is given) it reports a number of metrics related to Mahalanobis
    distances. Firstly it reports the average Mahalanobis distance of the
    snapshots in each .pcz file from their respective averages, and from
    each other's average structures. Then it reports on the percentage of
    snapshots in each .pcz file that lie within the 90% envelope of the
    snapshots in the other.
    """

    if args.verbosity:
        if args.verbosity > 1:
            log.basicConfig(format="%(levelname)s: %(message)s", 
                            level=log.DEBUG)
        elif args.verbosity == 1:
            log.basicConfig(format="%(levelname)s: %(message)s", level=log.INFO)
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")
     
    listLong = ['--input', '--quick', '--nvecs']
    listShort = ['-i']

    for i in range(1, len(sys.argv)):
        if (sys.argv[i] in listShort and 
            listLong[listShort.index(sys.argv[i])] in sys.argv):
            log.error('Please, use either the long or the short form of an '
                      + 'option but never both! Try again!')
            sys.exit(-1)

    if (args.input is None):
        log.error('')
        log.error('All or any of the mandatory command line arguments is '
                  + 'missing. The correct usage of pyPczcomp should be:')
        log.error('pyPczcomp -i|--input <input-file> [optional arguments]')
        log.error('')
        log.error('Type "pyPczcomp -h" or "pyPczcomp --help" for further' 
                  + ' details.')
        log.error('')
        sys.exit(-1)

    px = pca.load(args.input[0])
    py = pca.load(args.input[1])

    if px.n_atoms != py.n_atoms:
        print('Error: the number of atoms in the two files is different.')
        exit(1)

    if px.n_vecs < args.nvecs:
        print('Error: {0} only contains {1} vectors'.format(args.input[0],px.n_vecs))
        exit(1)

    if py.n_vecs < args.nvecs:
        print('Error: {0} only contains {1} vectors'.format(args.input[1],py.n_vecs))
        exit(1)

    n_atoms = px.n_atoms

    print('Comparison of X: {0} and Y: {1}'.format(args.input[0],args.input[1]))

    rms = alignment.rmsd_qcp(px.refxyz, py.refxyz)
    print('Rmsd between <X> and <Y>: {0:6.2f}'.format(rms))

    if not args.quick:
        print('Mahalanobis distances:')
        # mxx, myy, mxy and myx will contain the Mahalanobis distances of each snapshot
        # in X from <X>, of Y from <Y>, of X from <Y>, and of Y from <X>.
        mxx = np.zeros(px.n_frames)
        myy = np.zeros(py.n_frames)
        myx = np.zeros(py.n_frames)
        mxy = np.zeros(px.n_frames)
        evalx = px.evals
        evaly = py.evals
        for k in range(py.n_frames):
            xydif = py.frame(k)
            syy = py.scores(k)
            syx = px.map(xydif)
            smyy = 0.0
            smyx = 0.0
            for j in range(args.nvecs):
                smyx += syx[j]*syx[j]/evalx[j]
                smyy += syy[j]*syy[j]/evaly[j]
            myx[k] = np.sqrt(smyx)
            myy[k] = np.sqrt(smyy)

        for k in range(px.n_frames):
            xydif = px.frame(k)
            sxx = px.scores(k)
            sxy = py.map(xydif)
            smxx = 0.0
            smxy = 0.0
            for j in range(args.nvecs):
                smxy += sxy[j]*sxy[j]/evaly[j]
                smxx += sxx[j]*sxx[j]/evalx[j]
            mxy[k] = np.sqrt(smxy)
            mxx[k] = np.sqrt(smxx)

        print('X from <X>: {0:6.2f} +/- {1:6.2f}'.format(mxx.mean(),mxx.std()))
        print('Y from <Y>: {0:6.2f} +/- {1:6.2f}'.format(myy.mean(),myy.std()))
        print('X from <Y>: {0:6.2f} +/- {1:6.2f}'.format(mxy.mean(),mxy.std()))
        print('Y from <X>: {0:6.2f} +/- {1:6.2f}'.format(myx.mean(),myx.std()))

#    
# Calculation of the 90% enevlopes
#
        tx = np.sort(mxx)[int(0.9*px.n_frames)]
        ty = np.sort(myy)[int(0.9*py.n_frames)]

        i = 0
        for j in mxy:
            if j < tx:
                i += 1
        print('{0:3d}% of X lies within the 90% envelope of Y'.format(i*100//px.n_frames))

        i = 0
        for j in myx:
            if j < ty:
                i += 1
        print('{0:3d}% of Y lies within the 90% envelope of X'.format(i*100//py.n_frames))

#
# Dot product stuff. We rotate the eigenvectors in the second (Y) .pcz file
# using the rotation matrix found in least-squares fitting the average
# structures in the earlier RMSD calculation.
#
    evx = np.zeros((args.nvecs,3*n_atoms))
    evy = np.zeros((args.nvecs,3*n_atoms))
    T = alignment.compute_transformation(px.refxyz, py.refxyz)

    for i in range(args.nvecs):
        evx[i] = px.evecs[i]
        evy[i] = py.evecs[i]
        evy[i] = np.dot(evy[i].reshape(n_atoms,3),T.rotation).flatten()

    dp = np.zeros((args.nvecs,args.nvecs))
    ss = 0.0
    for i in range(args.nvecs):
        for j in range(args.nvecs):
            dp[i,j] = np.absolute(np.dot(evx[i],evy[j]))
            ss += dp[i,j]*dp[i,j]

    ss = ss/args.nvecs

    print('Dot product matrix:')
    print('    ', end=' ')
    for i in range(args.nvecs):
        print('{0:4d}'.format(i), end=' ')
    print(' ')
    for i in range(args.nvecs):
        print('{0:4d}'.format(i), end=' ')
        for j in range(args.nvecs):
            print('{0:4.2f}'.format(dp[i,j]), end=' ')
        print(' ')

    print('Subspace overlap: {0:6.3f}'.format(np.sqrt(ss)))
    amdp = (dp.max(axis=0).sum()+dp.max(axis=1).sum())/(2*args.nvecs)
    print('Average maximum dot product: {0:6.3f}'.format(amdp))
